/* Ramfile is based on truss.  calls an assembly module to install a tsr that
copies driver code and then stays resident
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <ctype.h>
#include <sys/stat.h>

#define DEFAULT_PAR 1024  /* = 16384 bytes, the size of an EMS page frame */     
                          /* we don't yet support EMS memory for the ramfile
                             but plan to do so later */
/* The EMS page frame sized buffer will be used as the ramfile if XMS memory
   is not available.   
*/

#define VERSION "2.1"  /* Change below too */

#define MOTTO "\n\t\t\tCHICKEN HAVEN SOFTWARE\n\
Ramfile (2.1), by Terry R. McConnell\n"

#define INFO  "\nramfile: install ramfile tsr\n\n\
with one ramfile for each listed path (mount point)\n\n\
usage:  ramfile [-h -v -d -u -x -a -q ] [<driver> [kbytes]] [path ...]\n\
   where -h[elp] or -? prints this message and exits\n\
         -v prints the version number and exits\n\
         -u uinstall the ramfile tsr and free memory \n\
	 -d use driver as name instead of devramx.bin \n\
         -q quiet(er) operation \n\
	 -a install at highest possible address (UMB if available)\n\
	 -x request kbytes of XMS memory instead of default 1024 \n\
	    if 0 then don't use XMS even if available \n\
    Ramfile without XMS is 16384 bytes in conventional memory\n"
		
#define MAX_ARG_LEN 16             /* Max length of combined options */
#define USAGE "Usage: ramfile [ -h -v -d -u -x -d -q -a ] [<driver.bin> [kbytes]] [path...]\n"

static char info[] = INFO;
unsigned char *code_buffer;
static unsigned char data_buffer[1024];
static char driver_name[64] = "devramx.bin";
int code_bytes;
static char doctored[16];

/* option processing code */
int find_opt(char *option_string);
int setcmp(const char *, const char *, char);

/* defined in truss.asm */
extern void setup(unsigned int par);
extern unsigned char quiet;
extern void uninstall(void);
extern void set_umb_link(void);
extern void restore_umb_link(void);
extern void go_tsr(void);
extern int strategy;
extern unsigned char LinkFlag;

unsigned MY_DS;  /* Data segment for non resident code = here */

/* entry point addresses stored by the tsr */
extern unsigned PSP_SEG;
extern unsigned driver_start; /* offsets from PSP_SEG: */
/* driver_payload used to be configurable from command line
but now we are making it compiled in */
static unsigned long driver_payload = DEFAULT_PAR;  

static unsigned xms_mem_k = 0x400;  

/* These jumps into driver code are computed below */
void (far *initialize_driver)(unsigned char, unsigned char, unsigned int); 

static int debug = 0;
static unsigned xflag = 0;
static unsigned dflag = 0;

/* Any time you add data to the top of device.asm this will need
   to be changed. Open device in debug.com and use dump (d) command
   to see how it starts. Better yet, don't add data to the top. */

unsigned char magic_number[3] = {0xEB,0x04,0x90};

struct device_data {
unsigned char magic[3];       /* First 3 bytes must be EB 30 00 */
unsigned char reserved[6];    /* 2 jump commands we ignore */
char names[16*MAX_FILES];
unsigned char termz;          
unsigned char nfiles;
} *my_device_data;

char mount_point[16];

int
main(int argc, char **argv)
{
	unsigned par = DEFAULT_PAR;
	int c=-1,n,i;                /* numeric option code */
	char cc;
	unsigned attributes;
	int nmounts = 0;
	char argstr[MAX_ARG_LEN] = "-";
	FILE *driver_file,*mpt;

	MY_DS = FP_SEG( (int far *)&debug ); /* just a trick to get our seg */  

	strategy = 0x00;  /* default is to camp at the lowest
				possible address */
	if(--argc > 0){  /* Anything on command line ? */
		if((*++argv)[0]=='-'){      /* already the end of options */
			if((*argv)[1]=='-'){
				++argv;
				c=-1;  /* no options */
			}
			else strcat(argstr,++*argv); /* start building argstr*/
			argc--;
			while((argc>0)&&((*++argv)[0]=='-')&&((*argv)[1]!='-')){
                   /* combine all options into a single string */
				argc--;
				strcat(argstr,++*argv);
			}
                        c=find_opt(argstr); /* look up code, set flags */
		}
		switch(c){

			case -1: break;  /* default, e.g no options */

			case 7: printf("%s\n",VERSION);
				return 0;
			case 8:
				dflag = 1;
				break;
			case 0: printf("%s\n",info);
				return 0;

			case 1: 
				set_umb_link();  
				uninstall();     /* This frees resident routine's memory */
				restore_umb_link();
				if(!quiet)
					printf("Uninstalled. Memory freed.\n");
				return 0;

			case 5:
				xflag = 1;
				break;
				

			case 3:  /* load high */
				strategy = 0x82;
				break;
                        case 2: 
                                break;
			case -2: ;  /* barf (don't use in your code) */
			default:
				fprintf(stderr,"%s: illegal option\n",argstr);
				return 1;
			}
	}
	if(argc > 0) {
		if(dflag) {
			strcpy(driver_name, *argv++);
			argc--;
		}
		driver_file = fopen(driver_name,"rb");
		if(!driver_file){
			fprintf(stderr,"unable to open driver %s\n",*argv);
			return 1;
		}
	}
	else {
		fprintf(stderr,"%s\n",USAGE);
		return 1;
	}
	if((argc > 0) && xflag) {
		xms_mem_k = atoi(*argv++);
		argc--;
	}
	if(!quiet) printf(MOTTO);
	
	/* see how big the driver is */
	n=0;
	while(fgetc(driver_file)!= EOF)n++;
	code_bytes = n;
	par = n/16;
	par++;   /* round up for good luck */
	par++;
	par += (unsigned)DEFAULT_PAR;  /* add on frame buffer */ 

        if(!quiet){

	         /* Print out information on what has happened */

		printf("Driver code %u paragraphs (%u bytes)\n",n/16,
			n);
		printf("Ramfile frame buffer size %lu paragraphs (%lu bytes)\n",
		driver_payload,
			driver_payload*16);
		printf("Total resident footprint: %u paragraphs (%u bytes)\n",par,
			par*16);
		printf("Use ramfile -u to uninstall and free memory\n");
       }

	/* copy to a buffer */
	code_buffer = (unsigned char *)malloc(n);
	if(!code_buffer){
		fprintf(stderr,"unable to allocate buffer for driver\n");
		fclose(driver_file);
		return 1;
	}

	rewind(driver_file); 
	for(i=0;i<n;i++)code_buffer[i]=((unsigned char)fgetc(driver_file));

	/* Check format of driver binary. This prevents a crash and burn
           if, e.g., we type device.asm instead of device.bin  */

	my_device_data = (struct device_data *)code_buffer;
	if((magic_number[0] != (my_device_data->magic)[0])
	   ||(magic_number[1] != (my_device_data->magic)[1])){
		printf("Bad device file format: no magic number\n");
		free(code_buffer);
		fclose(driver_file);
		return 1;
	}
	if(!argc){
		printf("At least one mount point required\n");
		return 1;
	}
	/* handle mount point names given on command line */
	while(argc){	
		if(nmounts >= MAX_FILES){
			printf("Too many mount points. Max = %d\n",MAX_FILES);
			return 1;
		}
		if((*argv)[1]==':') /* ignore drive letter if present */
			strcpy(doctored,(*argv)+2);
		else strcpy(doctored,*argv);
		if(strlen(doctored)> 15){
			printf("Mount point name %s too long\n",*argv);
			printf("Cannot exceed 15 characters\n");
			return 1;
		}/* normalize mountpoint names to upper case */
		for(i=0;i<strlen(doctored);i++)
			if(islower(doctored[i]))
				mount_point[i] = toupper(doctored[i]);
			else
				mount_point[i] = doctored[i];
		mount_point[i] = '\0';
		mpt = fopen(mount_point,"r");
		if(!mpt){
			printf("Cannot open mount point %s\n",mount_point);
			printf("A mount point exists and is read only.\n");
			printf("Make it so? (y or n)\n");
			cc = getchar();
			if(cc == 'y'){
				_dos_creat(mount_point,FA_RDONLY,(int *)mpt);
			}
			else	
			return 1;
		}	
		fclose(mpt);
		_dos_getfileattr(mount_point,&attributes);
		if(!(attributes & _A_RDONLY)){
			printf("A mount point is read only.\n");
			printf("Make it so? (y or n)\n");
			cc = getchar();
			if(cc == 'y'){
				attributes |= _A_RDONLY;
				_dos_setfileattr(mount_point,attributes);
			}
			else	
			 printf("You have been warned \n");
		}
		strcpy(my_device_data->names+nmounts*16,mount_point);
		nmounts++;
		if(!quiet)printf("Ramfile %d mount point name = %s\n",nmounts,mount_point);
		argv++;
		argc--;
	}
	my_device_data->nfiles = (unsigned char)nmounts;
        set_umb_link(); 
        setup(par);   /* call truss to set up memory resident code and data */

	if(debug)printf("Truss: code and data installed\n");

/* clean up */

	free(code_buffer);
	fclose(driver_file);

/* Tell the driver to initialize itself */
	if(debug)printf("Requesting %d bytes of extended memory\n",xms_mem_k*1024);

	initialize_driver = (void (far * )(unsigned char, unsigned char, unsigned int))MK_FP(PSP_SEG,driver_start);
/* first arg now ignored. Was mount_flag in version 1. could be
   resurrected for some other use */
	initialize_driver((unsigned char)1,(unsigned char)quiet, (unsigned int)xms_mem_k);
	if(debug)printf("Driver initialized  \n");

/* Leave driver installed */
	go_tsr(); 

        restore_umb_link(); 

	return 0; 
}

/*  find_opt: return a unique small integer for each possible option string */
/*  There should be a case in the switch statement in main to handle each. */
/*  -2 reserved for arg not found -1 reserved for no options */

int find_opt(char *word)
{
	if(strcmp(word,"-help")==0)return 0;
	if(strcmp(word,"-?")==0)return 0;
	if(strcmp(word,"-h")==0)return 0;
	if(strcmp(word,"-v")==0)return 7;
	if(strcmp(word,"-H")==0)return 0;
	if(strcmp(word,"-u")==0)return 1;
	if(strcmp(word,"-x")==0)return 5;
	if(strcmp(word,"-a")==0)return 3;
	if(strcmp(word,"-d")==0)return 8;
	if(strcmp(word,"-U")==0)return 1;
        if(strcmp(word,"-q")==0){
                quiet = 1;
                return 2;
        }
	if(setcmp(word,"-ax",'\1')){
		xflag = 1;
		return 3;
	}
	if(setcmp(word,"-dx",'\1')){
		xflag = 1;
		return 8;
	}
	
 	if(setcmp(word,"-qx",'\1')){
		quiet = 1;
		return 5;
	}
	if(setcmp(word,"-qax",'\1')){
		quiet = 1;
		xflag = 1;
		return 3;
	}
	if(setcmp(word,"-qdx",'\1')){
		quiet = 1;
		xflag = 1;
		return 8;
	}
	if(setcmp(word,"-adx",'\1')){
		dflag = 1;
		xflag = 1;
		return 3;
	}
	if(setcmp(word,"-qda",'\1')){
		quiet = 1;
		dflag = 1;
		return 3;
	}
	if(setcmp(word,"-qdax",'\1')){
		quiet = 1;
		dflag = 1;
		xflag = 1;
		return 3;
	}
        if(setcmp(word,"-qu",'\1')){
                quiet = 1;
                return 1;
        }
        if(setcmp(word,"-qa",'\1')){
                quiet = 1;
                return 3;
        }
/* args not found */
	return -2;
}


/* setcmp: string function that returns 1 if s and t are equal as sets,
           i.e., contain the same characters, not necessarily in the
           same order. Returns 0 if not, or if the answer cannot be
           determined. The invld character is any character not present
           in either s or t.  */

int setcmp(const char *s, const char *t, char invld)
{
        int i;
        char c;
        char *ptr;
        char buffer[MAX_ARG_LEN];

        if(s == NULL && t == NULL) return 1;       /* trivial case */
        if( t == NULL ) return 0; /* since here s must not be null */
        if(strchr(t,invld)!=NULL) {
                fprintf(stderr,"setcmp: error: character %c occurs in string %s\n",
                        invld,t);
               return 0;
        }
        if(strlen(t) != strlen(s)) return 0;

        /* From now on, both s and t are non-empty and of the same length */
        strcpy(buffer,t);
        for(i=0;i<strlen(s);i++){  /* loop over chars in s */
                c = *(s+i);  /* get next char of s */
                if((ptr=strchr(buffer,c))==NULL) return 0; /* No c in t */
                *ptr = invld;    /* else replace first occurence of c with
                                     invld so it won't be found again */
        }
        return 1; /* OK if we got to here */
}


